﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Text;
using System.Globalization;
using System.ComponentModel;

namespace MyControls {
    public class Geolocation : Control, INamingContainer {
        private const string DESIGN_TIME_HTML = "<table cellpadding=4 cellspacing=0 style=\"font:messagebox;color:buttontext;background-color:buttonface;border: solid 1px;border-top-color:buttonhighlight;border-left-color:buttonhighlight;border-bottom-color:buttonshadow;border-right-color:buttonshadow\"><tr><td nowrap><span style=\"font-weight:bold\">{0}</span> - {1}</td></tr><tr><td>{2}</td></tr></table>";
        private const string SCRIPT_CODE = @"<script type='text/javascript'>if(navigator.geolocation!=null){navigator.geolocation.getCurrentPosition(function(p){if(p!=null){document.getElementById('$field').value=['OK',p.timestamp,p.coords.latitude,p.coords.longitude,p.coords.accuracy,p.coords.altitude,p.coords.altitudeAccuracy,p.coords.heading,p.coords.speed].join(';');}},function(e){document.getElementById('$field').value=['ERR',e.code].join(';');},{enableHighAccuracy:$enableHighAccuracy,maximumAge:$maximumAge,timeout:$timeout});}</script>";
        private const string HIDDEN_FIELD_NAME = "GeoData";
        private const string HIDDEN_FIELD_VALUE = "ERR;-1";

        private HiddenField dataField;

        public enum GeolocationStatus {
            Unknown = -2,               // result unknown (no postback yet)
            Unsupported = -1,           // browser does not support geolocation or JS
            Success = 0,                // position obtained
            PermissionDenied = 1,       // user did not consent with geolocation
            PositionUnavailable = 2,    // position unavailable
            Timeout = 3,                // timeout
        }

        public Geolocation() {
            // Initialize deault values
            this.Status = GeolocationStatus.Unknown;
            this.EnableHighAccuracy = false;
            this.UpdateOnFirstPostbackOnly = true;
        }

        // Method overrides

        protected override void OnInit(EventArgs e) {
            base.OnInit(e);
            this.dataField = new HiddenField {
                ID = HIDDEN_FIELD_NAME,
                Value = HIDDEN_FIELD_VALUE
            };
            this.Controls.Add(this.dataField);
        }

        protected override void OnLoad(EventArgs e) {
            base.OnLoad(e);
            if (this.Page.IsPostBack) {
                var data = this.dataField.Value.Split(';');
                if (data[0].Equals("OK")) {
                    // Data obtained
                    this.Status = GeolocationStatus.Success;
                    this.Timestamp = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddMilliseconds(long.Parse(data[1])).ToLocalTime();
                    this.Latitude = ParseNullableDouble(data[2]);
                    this.Longitude = ParseNullableDouble(data[3]);
                    this.Accuracy = ParseNullableDouble(data[4]);
                    this.Altitude = ParseNullableDouble(data[5]);
                    this.AltitudeAccuracy = ParseNullableDouble(data[6]);
                    this.Heading = ParseNullableDouble(data[7]);
                    this.Speed = ParseNullableDouble(data[8]);
                }
                else {
                    // Error
                    this.Status = (GeolocationStatus)int.Parse(data[1]);
                }
            }
        }

        protected override void Render(HtmlTextWriter writer) {
            if (this.DesignMode) {
                // Show placeholder in design time
                writer.Write(DESIGN_TIME_HTML, "Geolocation", this.ID, "Server-side wrapper for Geolocation API.");
                return;
            }

            // Render hidden field
            base.Render(writer);

            if (!this.UpdateOnFirstPostbackOnly || !this.Page.IsPostBack) {
                var script = new StringBuilder(SCRIPT_CODE);
                script = script.Replace("$field", this.dataField.ClientID);
                script = script.Replace("$enableHighAccuracy", this.EnableHighAccuracy.ToString().ToLowerInvariant());
                script = script.Replace("$maximumAge", this.MaximumAgeString);
                script = script.Replace("$timeout", this.TimeoutString);
                writer.Write(script.ToString());
            }
        }

        // Configuration properties
        [Category("Options")]
        [Description("Specifies the maximum amount of time that user agent can use to obtain a position. When not set, the time is not limited.")]
        public TimeSpan? Timeout {
            get { return (TimeSpan?)this.ViewState["Timeout"]; }
            set { this.ViewState["Timeout"] = value; }
        }

        [Category("Options")]
        [Description("Indicates that the application is willing to accept a cached position whose age is no greater than the specified time. When omitted, the implementation must immediately attempt to acquire a new position object.")]
        public TimeSpan? MaximumAge {
            get { return (TimeSpan?)this.ViewState["MaximumAge"]; }
            set { this.ViewState["MaximumAge"] = value; }
        }

        [Category("Options"), DefaultValue(false)]
        [Description("When set to true, provides a hint that the application would like to receive the best possible results. This may result in slower response times or increased power consumption.")]
        public bool EnableHighAccuracy {
            get { return (bool)this.ViewState["EnableHighAccuracy"]; }
            set { this.ViewState["EnableHighAccuracy"] = value; }
        }

        [Category("Options"), DefaultValue(true)]
        [Description("When set to true, position is obtained only once, on first page load. When set to false, position is updated on every postback.")]
        public bool UpdateOnFirstPostbackOnly {
            get { return (bool)this.ViewState["UpdateOnFirstPostbackOnly"]; }
            set { this.ViewState["UpdateOnFirstPostbackOnly"] = value; }
        }

        // Output properties

        /// <summary>
        /// Gets the geolocation status.
        /// </summary>
        [Browsable(false)]
        public GeolocationStatus Status { get; private set; }

        /// <summary>
        /// Gets the time, when position was obtained, as reported by browser.
        /// </summary>
        [Browsable(false)]
        public DateTime Timestamp { get; private set; }

        /// <summary>
        /// Gets the latitude in decimal degrees.
        /// </summary>
        [Browsable(false)]
        public double? Latitude { get; private set; }

        /// <summary>
        /// Gets the longitude in decimal degrees.
        /// </summary>
        [Browsable(false)]
        public double? Longitude { get; private set; }

        /// <summary>
        /// Gets the browser-reported accuracy in meters.
        /// </summary>
        [Browsable(false)]
        public double? Accuracy { get; private set; }

        /// <summary>
        /// Gets the altitude in meters, according to WGS-84 reference model.
        /// </summary>
        [Browsable(false)]
        public double? Altitude { get; private set; }

        /// <summary>
        /// Gets the browser-reported altitude accuracy in meters.
        /// </summary>
        [Browsable(false)]
        public double? AltitudeAccuracy { get; private set; }

        /// <summary>
        /// Gets the heading (direction of movement) in degrees.
        /// </summary>
        [Browsable(false)]
        public double? Heading { get; private set; }

        /// <summary>
        /// Gets the movement speed in meters per second.
        /// </summary>
        [Browsable(false)]
        public double? Speed { get; private set; }

        // Private helpers

        private static double? ParseNullableDouble(string s) {
            if (string.IsNullOrWhiteSpace(s)) return null;
            double d;
            if (!double.TryParse(s, NumberStyles.Any, CultureInfo.InvariantCulture, out d)) return null;
            return d;
        }

        private string TimeoutString {
            get {
                if (!this.Timeout.HasValue) return "Infinity";
                return Timeout.Value.TotalMilliseconds.ToString(CultureInfo.InvariantCulture);
            }
        }

        private string MaximumAgeString {
            get {
                if (!this.MaximumAge.HasValue) return "0";
                return Timeout.Value.TotalMilliseconds.ToString(CultureInfo.InvariantCulture);
            }
        }

    }
}